from __future__ import annotations

from typing import Any, List


def add_str_trail(string: str | bytes) -> str | bytes:
    """
    Appends the null character b"\x00" to the end of the bytes if it is not already present

    Args:
        string: The string/bytes to append the null character to

    Returns:
        Bytes ending with a null character
    """
    if not has_str_trail(string):
        string += b"\x00"
    return string


def has_str_trail(string: str | bytes) -> bool:
    """
    Checks if the given bytes ends with the null character

    Args:
        string: The string/bytes to check the last character of

    Returns:
        `True` if the given bytes end with a null character else 0
    """
    return string.endswith(b"\x00")


def del_str_trail(string: str | bytes) -> str | bytes:
    """
    Removes the the null character from the end of a string/bytes

    Args:
        string: The string/bytes to remove the null character from

    Returns:
        string/bytes with the null character removed from the end
    """
    if has_str_trail(string):
        string = string[:-1]
    return string


def add_prefix_chars(string: str, char: str, length: int) -> str:
    """
    Adds the specified character at the front of a string so that the string is of a given length

    Args:
        string: The string to prepend the characters to
        char: The character to prepend
        length: The desired length of the final string

    Returns:
        A string with the required number of characters prepended to make it of the given length
    """
    if len(string) > length:
        return string
    else:
        return char * (length - len(string)) + string


def add_suffix_chars(string: str, char: str, total_length: int) -> str:
    """
    Adds the specified character at the end of a string so that the string is of a given length

    Args:
        string: The string to append the characters to
        char: The character to append
        total_length: The desired length of the final string

    Returns:
        A string with the required number of characters appended to make it of the given length
    """
    if len(string) > total_length:
        return string
    else:
        return string + char * (total_length - len(string))


def remove_prefix(string: str, prefix: str) -> str:
    """
    Cheap knockoff function of: https://docs.python.org/3/library/stdtypes.html?highlight=removesuffix#str.removeprefix

    Args:
        string: The string to check it's prefix
        prefix: The prefix to remove from the string

    Returns:
        The given string or the string without the prefix if it was present
    """
    return string[len(prefix):] if string.index(prefix) == 0 else string


def q_str(value: Any) -> str:
    """
    Converts the given value into a string. If the value is in string or byte form, quotes are added
    around the value. Bytes are also represented in hex.

    Args:
        value: The value to return the string representation of

    Returns:
        The string representation of the given value. Slightly modified if it's a string or bytes value.
    """
    if type(value) is str:
        return f"'{value}'"
    if type(value) is bytes:
        return f"b'{value.hex()}'"
    return str(value)


def trunc_string(string: str | bytes, length=30, add_ellipsis=True) -> str | bytes:
    """
    Truncates string to a certain length and possibly adds ellipsis to said string.

    Args:
        string: The string to truncate if its length is longer than the specified max (see length)
        length: The max given length the string can be
        add_ellipsis: If ellipsis should be added if the string is truncated

    Returns:
        The truncated string (with ellipsis) if it was longer than the given length
    """
    ellipsis_ = '...' if isinstance(string, str) else b'...'
    alt = '' if isinstance(string, str) else b''
    return (string[:length] + (ellipsis_ if add_ellipsis else alt)) if len(string) > length else string


def add_tabs(string: str, tabs: int) -> str:
    """
    Adds tabs to a string. Tabs are added in front of each line of the string

    Args:
        string: The string to add tabs to
        tabs: The amount of tabs to add before each line

    Returns:
        The tabbed string
    """
    split_string = string.splitlines(keepends=True)
    return ("\t" * tabs) + ("\t" * tabs).join(split_string) 


def create_inline_line(entries: List[Any]):
    """
    Creates inline string of elements from the given list, converts the elements to strings and appends them.
    The created string always starts with a tab and ends with a newline.

    Args:
        entries: The entries to convert to an inline, comma separated, string

    Returns:
        The created inline string
    """
    return "\t" + ", ".join(map(str, entries)) + "\r\n"


def create_textual_hex(string: str, space_distance: int = 2, enter_distance: int = 48):
    """
    Converts string to a textual hex format. Splitting the string with spaces and enters spaced appropriately.

    **Please note**: that the 'enter_distance' parameter is including the - to be added - spaces. If you calculated it
    without the spaces, please multiply the number by: `block size incl space / block size excl space`

    Args:
        string: The string to split up
        space_distance: The distance between spaces. Defaults to 2.
        enter_distance: The distance between enters (includes added spaces). Defaults to 48 (Every 32 string chars).

    Returns:
        The converted string with spaces and enters
    """
    return insert_char(insert_char(string, " ", space_distance), "\n", enter_distance)


def insert_char(string: str, char: str, step: int=64):
    """
    Inserts a character every x amount of letters into a string.

    Args:
        string: The string to add the letters into
        char: The character to add to the string
        step: The amount of characters in between each character insertion

    Returns:
        The created string with the new characters within it

    :Author:
        gurney alex @ https://stackoverflow.com/a/2657733/7230293
    """
    return char.join(string[i:i + step] for i in range(0, len(string), step))



from __future__ import annotations

from enum import Enum
from typing import Dict, List, TYPE_CHECKING
from uuid import UUID

from AoE2ScenarioParser.helper import bytes_parser
from AoE2ScenarioParser.helper.list_functions import listify
from AoE2ScenarioParser.helper.string_manipulations import create_textual_hex, insert_char, add_suffix_chars, q_str, \
    trunc_string
from AoE2ScenarioParser.sections.aoe2_struct_model import AoE2StructModel, model_dict_from_structure
from AoE2ScenarioParser.sections.dependencies.dependency import handle_retriever_dependency
from AoE2ScenarioParser.sections.retrievers.retriever import Retriever, reset_retriever_map

if TYPE_CHECKING:
    from AoE2ScenarioParser.helper.incremental_generator import IncrementalGenerator


class SectionLevel(Enum):
    """This enum class indicates the section level of the structure"""
    TOP_LEVEL = 0
    STRUCT = 1


class SectionName(Enum):
    """This enum class indicates the different sections in the scenario file"""
    FILE_HEADER = "FileHeader"
    DATA_HEADER = "DataHeader"
    MESSAGES = "Messages"
    CINEMATICS = "Cinematics"
    BACKGROUND_IMAGE = "BackgroundImage"
    PLAYER_DATA_TWO = "PlayerDataTwo"
    GLOBAL_VICTORY = "GlobalVictory"
    DIPLOMACY = "Diplomacy"
    OPTIONS = "Options"
    MAP = "Map"
    UNITS = "Units"
    TRIGGERS = "Triggers"
    FILES = "Files"


class AoE2FileSection:
    """
    Multiple retrievers and structures containing related data are grouped together under one file section. This class
    is used to construct, manage and commit all the retrievers and structures inside a file section.

    Utilises the AoE2StructModel to create multiple of the same structure. These copies are what is used to store the
    data from a scenario file
    """

    def __init__(
            self,
            name: str,
            retriever_map: Dict[str, 'Retriever'],
            uuid: UUID,
            struct_models: Dict[str, 'AoE2StructModel'] = None,
            level: SectionLevel = SectionLevel.TOP_LEVEL
    ):
        """
        Args:
            name: The name of the file section
            retriever_map: A dict containing all the retrievers for this file section
            uuid: The universally unique identifier for the scenario that this file section belongs to
            struct_models: A dict containing all the structures for this file section
            level: This variable tells if a structure or file section is at the top level of the scenario or a struct.
        """
        if struct_models is None:
            struct_models = {}

        self.name: str = name
        self.retriever_map: Dict[str, 'Retriever'] = retriever_map
        self._uuid = uuid
        self.byte_length: int = -1
        self.struct_models: Dict[str, AoE2StructModel] = struct_models
        self.level: SectionLevel = level

    def find_struct_model_by_retriever(self, retriever: 'Retriever') -> AoE2StructModel:
        prefix = "struct:"

        struct_datatype = retriever.datatype.var
        if not struct_datatype.startswith(prefix):
            raise ValueError(f"Unable to retrieve model '{struct_datatype}' from file section. "
                             f"Possible model names: {self.struct_models.keys()}")

        struct_name = struct_datatype[len(prefix):]
        return self.struct_models[struct_name]

    @classmethod
    def from_model(cls, model: AoE2StructModel, uuid: UUID, set_defaults: bool = False) -> AoE2FileSection:
        """
        Create a file section based on the given AoE2StructModel.

        Args:
            model: The model to copy from
            uuid: The universally unique identifier of the scenario
            set_defaults: If retrievers need to be set to the default values

        Returns:
            An AoE2FileSection instance based on the model
        """
        duplicate_rmap = {k: e.duplicate() for k, e in model.retriever_map.items()}

        if set_defaults:
            reset_retriever_map(duplicate_rmap)

        return cls(
            name=model.name,
            retriever_map=duplicate_rmap,
            uuid=uuid,
            struct_models=model.structs,
            level=SectionLevel.STRUCT
        )

    @classmethod
    def from_structure(cls, section_name: str, structure: Dict[str, Dict], uuid: UUID) -> 'AoE2FileSection':
        """
        Constructs all retrievers and structures from the given file section (dict)

        Args:
            section_name (str): This is the name of the file section
            structure (Dict[str, Dict]): The dict representation of the file section to build
            uuid (UUID): The universally unique identifier of the scenario that this file section belongs to

        Returns:
            An AoE2FileSection instance representing the given file section structure
        """
        retriever_map = {}
        log_all_retrievers = structure.get('log', False)
        for name, attr in structure.get('retrievers').items():
            if log_all_retrievers:
                attr['log'] = True
            retriever_map[name] = Retriever.from_structure(name, attr)

        structs = model_dict_from_structure(structure)
        return cls(section_name, retriever_map, uuid, structs)

    def get_data_as_bytes(self) -> bytes:
        """
        Converts the data of a file section into bytes

        Returns:
            The bytes representing the file section
        """
        result: List[bytes] = []
        for retriever in self.retriever_map.values():
            result.append(retriever.get_data_as_bytes())
        return b''.join(result)

    def set_data_from_generator(self, igenerator: 'IncrementalGenerator') -> None:
        """
        Sets all the data for the retrievers and structures of the file section from the given generator
        (bytes of the scenario file). Each retriever (and retrieves inside structures) parses a set number of bytes
        based on its datatype. The total number of bytes parsed is then stored in the self.bytes_length variable

        Args:
            igenerator: A generator from a binary scenario file

        Raises:
            ValueError: if a structure inside a model is not defined
        """
        total_length = 0
        for retriever in self.retriever_map.values():
            handle_retriever_dependency(retriever, "construct", self, self._uuid)
            if retriever.datatype.type == "struct":
                struct_name = retriever.datatype.get_struct_name()

                structs = []
                for _ in range(retriever.datatype.repeat):
                    model = self.struct_models.get(struct_name)
                    if model is None:
                        raise ValueError(f"Model '{struct_name}' not found. Likely not defined in structure.")

                    struct = self._create_struct(model, igenerator)
                    structs.append(struct)

                    total_length += struct.byte_length
                retriever.set_data(structs, affect_dirty=False)
            else:
                retrieved_bytes = bytes_parser.retrieve_bytes(igenerator, retriever)
                self._fill_retriever_with_bytes(retriever, retrieved_bytes)

                total_length += sum([len(raw_bytes) for raw_bytes in retrieved_bytes])

        self.byte_length = total_length

    def _fill_retriever_with_bytes(self, retriever: 'Retriever', retrieved_bytes: List[bytes]) -> None:
        """
        With the given bytes, sets the data of the provided retriever.

        If setting the data fails and a ValueError is raised, this function deals with the error logging (and re-raises
        the error).

        Args:
            retriever: The retriever to set the data for
            retrieved_bytes: The bytes to set the data from

        Raises:
            ValueError: if the bytes given cannot be used to set the data of the retriever provided
        """
        try:
            retriever.set_data_from_bytes(retrieved_bytes)
        except ValueError as e:
            print("\n\n" + "#" * 120)
            print(f"[{e.__class__.__name__}] Occurred while setting data in:\n\t{retriever}")
            print("#" * 120)
            print(trunc_string(self.get_byte_structure_as_string(), length=10000))
            raise e

    def _create_struct(self, model: 'AoE2StructModel', igenerator: 'IncrementalGenerator') -> 'AoE2FileSection':
        """
        Creates an AoE2FileSection from the given model using data from the provided generator

        If setting the data fails and a ValueError is raised, this function deals with the error logging (and re-raises
        the error).

        Args:
            model: The model to make the AoE2FileSection from
            igenerator: The generator to set the data for the file section from

        Returns:
            An AoE2FileSection object created from the given model with its data set according to the given generator
        """
        struct = AoE2FileSection.from_model(model, uuid=self._uuid)

        try:
            struct.set_data_from_generator(igenerator)
        except (TypeError, ValueError) as e:
            print("\n" + "#" * 120)
            print(f"[{e.__class__.__name__}] Occurred while setting data in: {struct.name}"
                  f"\n\tContent of {self.name}:")
            print("#" * 120)
            print(trunc_string(self.get_byte_structure_as_string(), length=10000))
            raise e

        return struct

    def __getattr__(self, item):
        """Providing a default way to access retriever data labeled 'name'"""
        if 'retriever_map' not in self.__dict__:
            return super().__getattribute__(item)
        elif item.startswith("__"):
            return super().__getattribute__(item)
        else:
            retriever = self.retriever_map[item]
            if retriever is None:
                return super().__getattribute__(item)
            else:
                return retriever.data

    def __setattr__(self, name, value):
        """Trying to edit retriever data labeled 'name' if available"""
        if 'retriever_map' not in self.__dict__:
            super().__setattr__(name, value)
        else:
            try:
                self.retriever_map[name].data = value
            except KeyError:
                super().__setattr__(name, value)

    def _entry_to_string(self, name: str, data: str, datatype: str) -> str:
        """
        Formats the given data into a string. Example::

            version: 1.45 (1 * c4)

        Args:
            name: The name of the value
            data: The value itself
            datatype: The datatype of the value

        Returns:
            Formatted string representation of the given data
        """
        prefix = "\t"
        if self.level == SectionLevel.STRUCT:
            prefix = "\t\t\t"
        if 'str' in datatype:
            data = q_str(data)
        return f"{prefix}{name}: {data} ({datatype})\n"

    def get_header_string(self) -> str:
        """
        Returns:
            The name of the header for the file section
        """
        if self.level == SectionLevel.TOP_LEVEL:
            return f"######################## {self.name} ######################## [SECTION]"
        elif self.level == SectionLevel.STRUCT:
            return f"############ {self.name} ############  [STRUCT]"

    def get_byte_structure_as_string(self):
        """
        Recursively converts the retrievers of a file section and their data into a readable format. An
        example with this format is shown below::

            ########################### units (1954 * struct:UnitStruct)
            ############ UnitStruct ############  [STRUCT]
            00 00 70 42                 x (1 * f32): 60.0
            00 00 70 42                 y (1 * f32): 60.0
            00 00 00 00                 z (1 * f32): 0.0
            52 05 00 00                 reference_id (1 * s32): 1362
            89 02                       unit_const (1 * u16): 649
            02                          status (1 * u8): 2
            00 00 00 00                 rotation (1 * f32): 0.0
            00 00                       initial_animation_frame (1 * u16): 0
            ff ff ff ff                 garrisoned_in_id (1 * s32): -1

        Returns:
            A string showing the data in each retriever of the file section
        """
        byte_structure = "\n" + self.get_header_string()

        for key, retriever in self.retriever_map.items():
            byte_structure += "\n"

            listed_retriever_data = listify(retriever.data)
            struct_header_set = False
            for struct in listed_retriever_data:
                if isinstance(struct, AoE2FileSection):
                    if not struct_header_set:
                        byte_structure += f"\n{'#' * 27} {key} ({retriever.datatype.to_simple_string()})"
                        struct_header_set = True
                    byte_structure += struct.get_byte_structure_as_string()
            # Struct Header was set. Retriever was struct, data retrieved using recursion. Next retriever.
            if struct_header_set:
                byte_structure += f"{'#' * 27} End of: {key} ({retriever.datatype.to_simple_string()})\n"
                continue

            retriever_data_bytes = retriever.get_data_as_bytes().hex()
            retriever_hex = create_textual_hex(retriever_data_bytes, space_distance=2, enter_distance=24)

            split_hex = retriever_hex.split("\n")
            split_hex_length = len(split_hex)

            retriever_short_string: str = retriever.get_short_str()
            split_data_string = retriever_short_string.replace('\x00', ' ').splitlines()
            data_lines = []
            for x in split_data_string:
                if len(x) > 120:
                    data_lines += [f'\t{x}' for x in insert_char(x, '\r\n', 120).splitlines()]
                else:
                    data_lines.append(x)
            split_data_length = len(data_lines)

            lines = max(split_hex_length, split_data_length)

            combined_strings = []
            for i in range(0, lines):
                hex_part = split_hex[i] if i < split_hex_length else ""
                data_part = data_lines[i] if i < split_data_length else ""
                combined_strings.append(add_suffix_chars(hex_part, " ", 28) + data_part)

            byte_structure += "\n".join(combined_strings)

        return byte_structure + "\n"

    def __str__(self):
        represent = self.name + ": \n"

        for retriever in self.retriever_map.values():
            if type(retriever.data) is list and len(retriever.data) > 0:
                if isinstance(retriever.data[0], AoE2FileSection):
                    represent += "\t" + retriever.name + ": [\n"
                    for x in retriever.data:
                        represent += "\t\t" + str(x)
                    represent += "\t]\n"
                else:
                    represent += self._entry_to_string(
                        retriever.name,
                        str(retriever.data),
                        str(retriever.datatype.to_simple_string())
                    )
            else:
                data = retriever.data or "None"
                represent += self._entry_to_string(
                    retriever.name, str(data), str(retriever.datatype.to_simple_string())
                )

        return represent

    def __repr__(self):
        return f"<AoE2FileSection> {self.name}"



from typing import TypedDict

import requests

from common import tab, train_case


UseDesc = str
FlagValue = str
FlagDesc = str
class AttrDesc(TypedDict):
    name: str
    desc: str
    notes: list[UseDesc]
    flags: dict[FlagValue, FlagDesc]

AttrId = int
AttrDescs = dict[AttrId, AttrDesc]

def gen_docstr(desc: AttrDesc) -> str:
    docstr = [
        tab('"' * 3),
        desc['desc'],
    ]

    if len(desc['flags']) > 0:
        docstr.extend([
            "",
            "- Flags:",
            "",
        ])

        for val, use in desc['flags'].items():
            docstr.append(
                tab(f"- {val}: {use}")
            )

    if len(desc['notes']) > 0:
        docstr.extend([
            "",
            "- Notes:",
            "",
            *map(tab, desc["notes"]),
        ])

    docstr.append('"'*3)
    return "\n    ".join(docstr)

def gen_attr_class(attrs: AttrDescs) -> str:
    klass = [r"class ObjectAttribute(_DataSetIntEnums):"]

    for id_, desc in attrs.items():
        klass.extend([
            tab(train_case(desc["name"]) + f" = {id_}"),
            gen_docstr(desc)
        ])

    return "\n".join(klass)


def main():
    res = requests.get(r"https://raw.githubusercontent.com/Divy1211/AoE2DE_UGC_Guide/main/docs/general/attributes/attributes.json")
    attrs = res.json()

    with open("attributes.py", "w") as file:
        file.write(gen_attr_class(attrs))


if __name__ == "__main__":
    main()



from typing import List, Dict, Any, TYPE_CHECKING, Type
from uuid import UUID

from AoE2ScenarioParser.exceptions.asp_exceptions import ScenarioWritingError
from AoE2ScenarioParser.objects.support.uuid_list import NO_UUID
from AoE2ScenarioParser.sections.retrievers.construct_progress import ConstructProgress
from AoE2ScenarioParser.sections.retrievers.retriever_object_link_parent import RetrieverObjectLinkParent

if TYPE_CHECKING:
    from AoE2ScenarioParser.objects.aoe2_object import AoE2Object
    from AoE2ScenarioParser.sections.retrievers.retriever_object_link import RetrieverObjectLink


class RetrieverObjectLinkGroup(RetrieverObjectLinkParent):
    def __init__(
            self,
            section_name: str,
            link: str = "",
            group: List['RetrieverObjectLink'] = None
    ):
        super().__init__(section_name, link)

        self.group = group or []

        # Inject reference to the parent into the child for debugging purposes
        for rol in self.group:
            rol.parent = self

    def get_names(self):
        return [e.name for e in self.group]

    def pull(
            self,
            uuid: UUID,
            number_hist: List[int] = None,
            host_obj: Type['AoE2Object'] = None,
            progress: ConstructProgress = None
    ) -> Dict[str, Any]:
        if number_hist is None:
            number_hist = []

        attributes = {}

        section = self.pull_from_link(uuid, number_hist, progress=progress)

        # 'done' is set to 0 as entries within a group do not have a prefix that has already been cleared
        progress = ConstructProgress(section=section, done=0)

        for link in self.group:
            attributes[link.name] = link.pull_from_link(uuid=uuid, number_hist=number_hist,
                                                        host_obj=host_obj, progress=progress)
        return attributes

    def push(self, uuid: UUID, host_obj: 'AoE2Object'):
        if uuid == NO_UUID:
            raise ValueError(f"Invalid object push. No UUID was set. Object class: {host_obj.__class__.__name__}")

        number_hist = host_obj.instance_number_history
        section = self.pull_from_link(uuid, number_hist)

        try:
            for link in reversed(self.group):
                link.push_to_link(
                    uuid=uuid, number_hist=number_hist, host_obj=host_obj,
                    # 'done' is set to 0 as entries within a group do not have a prefix that has already been cleared
                    progress=ConstructProgress(section=section, done=0)
                )
        except Exception as e:
            if isinstance(e, ScenarioWritingError):
                raise e

            raise ScenarioWritingError('\n\n' + '\n'.join([
                "An error was raised while trying to write the scenario.",
                "The error occurred on the following location:",
                "\tSection:  " + self.section_name,
                "\tLocation: " + self.format_link_string(self.link, number_hist),
                "",
                "You can see the original error at the top of the stack trace."
            ]))
